/*
 * XdccDownloader - Connection.java - Copyright © 2013 David Roden
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.pterodactylus.irc;

import static com.google.common.base.Preconditions.checkState;
import static java.util.concurrent.TimeUnit.SECONDS;

import java.io.BufferedReader;
import java.io.Closeable;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.net.SocketFactory;

import net.pterodactylus.irc.connection.ChannelNotJoinedHandler;
import net.pterodactylus.irc.connection.ConnectionEstablishHandler;
import net.pterodactylus.irc.event.ChannelJoined;
import net.pterodactylus.irc.event.ChannelLeft;
import net.pterodactylus.irc.event.ChannelMessageReceived;
import net.pterodactylus.irc.event.ChannelNicknames;
import net.pterodactylus.irc.event.ChannelNotJoined;
import net.pterodactylus.irc.event.ChannelNotJoined.Reason;
import net.pterodactylus.irc.event.ChannelNoticeReceived;
import net.pterodactylus.irc.event.ChannelTopic;
import net.pterodactylus.irc.event.ClientQuit;
import net.pterodactylus.irc.event.ConnectionClosed;
import net.pterodactylus.irc.event.ConnectionEstablished;
import net.pterodactylus.irc.event.ConnectionFailed;
import net.pterodactylus.irc.event.DccAcceptReceived;
import net.pterodactylus.irc.event.DccSendReceived;
import net.pterodactylus.irc.event.KickedFromChannel;
import net.pterodactylus.irc.event.MotdReceived;
import net.pterodactylus.irc.event.NicknameChanged;
import net.pterodactylus.irc.event.NicknameInUseReceived;
import net.pterodactylus.irc.event.NoNicknameGivenReceived;
import net.pterodactylus.irc.event.PrivateMessageReceived;
import net.pterodactylus.irc.event.PrivateNoticeReceived;
import net.pterodactylus.irc.event.ReplyReceived;
import net.pterodactylus.irc.event.UnknownReplyReceived;
import net.pterodactylus.irc.util.RandomNickname;
import net.pterodactylus.xdcc.util.io.BandwidthCountingInputStream;
import net.pterodactylus.xdcc.util.io.BandwidthCountingOutputStream;

import com.google.common.base.Optional;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.google.common.eventbus.EventBus;
import com.google.common.eventbus.Subscribe;
import com.google.common.io.Closeables;
import com.google.common.primitives.Ints;
import com.google.common.primitives.Longs;
import com.google.common.util.concurrent.AbstractExecutionThreadService;
import com.google.common.util.concurrent.Service;
import org.apache.log4j.Logger;

/**
 * A connection to an IRC server.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class Connection extends AbstractExecutionThreadService implements Service {

	/* The logger. */
	private static final Logger logger = Logger.getLogger(Connection.class.getName());

	/** The event bus. */
	private final EventBus eventBus;

	/** The socket factory. */
	private final SocketFactory socketFactory;

	/** The hostname to connect to. */
	private final String hostname;

	/** The port to connect to. */
	private final int port;

	/** The nickname chooser. */
	private NicknameChooser nicknameChooser = new NicknameChooser() {

		@Override
		public String getNickname() {
			return RandomNickname.get();
		}
	};

	/** The nickname. */
	private String nickname = null;

	/** The username. */
	private Optional<String> username = Optional.absent();

	/** The real name. */
	private Optional<String> realName = Optional.absent();

	/** The optional password for the connection. */
	private Optional<String> password = Optional.absent();

	/** The connection handler. */
	private ConnectionHandler connectionHandler;

	/** Whether the connection has already been established. */
	private final AtomicBoolean established = new AtomicBoolean();

	/**
	 * Creates a new connection.
	 *
	 * @param eventBus
	 * 		The event bus
	 * @param socketFactory
	 * 		The socket factory
	 * @param hostname
	 * 		The hostname of the IRC server
	 * @param port
	 * 		The port number of the IRC server
	 */
	public Connection(EventBus eventBus, SocketFactory socketFactory, String hostname, int port) {
		this.eventBus = eventBus;
		this.socketFactory = socketFactory;
		this.hostname = hostname;
		this.port = port;
	}

	//
	// ACCESSORS
	//

	/**
	 * Returns the hostname of the remote end of the connection.
	 *
	 * @return The remote’s hostname
	 */
	public String hostname() {
		return hostname;
	}

	/**
	 * Returns the port number of the remote end of the connection.
	 *
	 * @return The remote’s port number
	 */
	public int port() {
		return port;
	}

	/**
	 * Returns whether this connection has already been established.
	 *
	 * @return {@code true} as long as this connection is established, {@code
	 *         false} otherwise
	 */
	public boolean established() {
		return established.get();
	}

	/**
	 * Returns the nickname that is currently in use by this connection. The
	 * nickname is only available once the connection has been {@link #start()}ed.
	 *
	 * @return The current nickname
	 */
	public String nickname() {
		return nickname;
	}

	//
	// MUTATORS
	//

	/**
	 * Sets the nickname chooser. The nickname chooser is only used during the
	 * creation of the connection.
	 *
	 * @param nicknameChooser
	 * 		The nickname chooser
	 * @return This connection
	 */
	public Connection nicknameChooser(NicknameChooser nicknameChooser) {
		this.nicknameChooser = nicknameChooser;
		return this;
	}

	/**
	 * Sets the username to use.
	 *
	 * @param username
	 * 		The username to use
	 * @return This connection
	 */
	public Connection username(String username) {
		this.username = Optional.fromNullable(username);
		return this;
	}

	/**
	 * Sets the real name to use.
	 *
	 * @param realName
	 * 		The real name to use
	 * @return This connection
	 */
	public Connection realName(String realName) {
		this.realName = Optional.fromNullable(realName);
		return this;
	}

	/**
	 * Sets the optional password for the connection.
	 *
	 * @param password
	 * 		The password for the connection
	 * @return This connection
	 */
	public Connection password(String password) {
		this.password = Optional.fromNullable(password);
		return this;
	}

	//
	// ACTIONS
	//

	/**
	 * Returns the current rate of the connection’s incoming side.
	 *
	 * @return The current input rate (in bytes per second)
	 */
	public long getInputRate() {
		return (connectionHandler != null) ? connectionHandler.getInputRate() : 0;
	}

	/**
	 * Returns the current rate of the connection’s outgoing side.
	 *
	 * @return The current output rate (in bytes per second)
	 */
	public long getOutputRate() {
		return (connectionHandler != null) ? connectionHandler.getOutputRate() : 0;
	}

	/**
	 * Checks whether the given source is the client represented by this
	 * connection.
	 *
	 * @param source
	 * 		The source to check
	 * @return {@code true} if this connection represents the given source, {@code
	 *         false} otherwise
	 */
	public boolean isSource(Source source) {
		return source.nick().isPresent() && source.nick().get().equals(nickname);
	}

	/**
	 * Joins the given channel.
	 *
	 * @param channel
	 * 		The channel to join
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	public void joinChannel(final String channel) throws IOException {
		connectionHandler.sendCommand("JOIN", channel);
	}

	/**
	 * Sends a message to the given recipient, which may be a channel or another
	 * nickname.
	 *
	 * @param recipient
	 * 		The recipient of the message
	 * @param message
	 * 		The message
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	public void sendMessage(String recipient, String message) throws IOException {
		connectionHandler.sendCommand("PRIVMSG", recipient, message);
	}

	/**
	 * Sends a DCC RESUME request to the given recipient.
	 *
	 * @param recipient
	 * 		The recipient of the request
	 * @param filename
	 * 		The name of the file to resume
	 * @param port
	 * 		The port number from the original DCC SEND request
	 * @param position
	 * 		The position at which to resume the transfer
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	public void sendDccResume(String recipient, String filename, int port, long position) throws IOException {
		connectionHandler.sendCommand("PRIVMSG", recipient, String.format("\u0001DCC RESUME %s %d %d\u0001", filename, port, position));
	}

	/**
	 * Closes this connection.
	 *
	 * @throws IOException
	 * 		if an I/O error occurs
	 */
	public void close() throws IOException {
		if (connectionHandler != null) {
			connectionHandler.close();
		}
	}

	//
	// ABSTRACTEXECUTIONTHREADSERVICE METHODS
	//

	@Override
	protected void startUp() throws IllegalStateException {
		checkState(username.isPresent(), "username must be set");
		checkState(realName.isPresent(), "realName must be set");
	}

	@Override
	protected void run() {

		/* connect to remote socket. */
		try {
			Socket socket = socketFactory.createSocket(hostname, port);
			socket.setSoTimeout((int) TimeUnit.MINUTES.toMillis(3));
			connectionHandler = new ConnectionHandler(socket.getInputStream(), socket.getOutputStream());

			/* register connection. */
			if (password.isPresent()) {
				connectionHandler.sendCommand("PASSWORD", password.get());
			}
			connectionHandler.sendCommand("USER", username.get(), "8", "*", realName.get());
			nickname = nicknameChooser.getNickname();
			connectionHandler.sendCommand("NICK", nickname);

		} catch (IOException ioe1) {
			eventBus.post(new ConnectionFailed(this, ioe1));
			return;
		}

		eventBus.register(this);
		/* now read replies and react. */
		try {
			/* some status variables. */
			int oldConnectionStatus = 0;
			int connectionStatus = 0;
			boolean connected = true;
			StringBuilder motd = new StringBuilder();
			Set<Nickname> nicks = Sets.newHashSet();

			/* server modes. */
			Map<String, String> nickPrefixes = Maps.newHashMap();
			Set<Character> channelTypes = Sets.newHashSet();

			ConnectionEstablishHandler connectionEstablishHandler = new ConnectionEstablishHandler(eventBus, this);
			ChannelNotJoinedHandler channelNotJoinedHandler = new ChannelNotJoinedHandler(eventBus, this);

			while (connected) {
				Reply reply = connectionHandler.readReply();
				eventBus.post(new ReplyReceived(this, reply));
				logger.trace(String.format("<< %s", reply));
				String command = reply.command();
				List<String> parameters = reply.parameters();

				/* most common events. */
				if (command.equalsIgnoreCase("PRIVMSG")) {
					String recipient = parameters.get(0);
					String message = parameters.get(1);
					if (message.startsWith("\u0001") && message.endsWith("\u0001")) {
						/* CTCP! */
						handleCtcp(reply.source().get(), message);
					} else if (!channelTypes.contains(recipient.charAt(0))) {
						eventBus.post(new PrivateMessageReceived(this, reply.source().get(), message));
					} else {
						eventBus.post(new ChannelMessageReceived(this, recipient, reply.source().get(), message));
					}

				} else if (command.equalsIgnoreCase("NOTICE")) {
					String recipient = parameters.get(0);
					String message = parameters.get(1);
					if (message.startsWith("\u0001") && message.endsWith("\u0001")) {
						/* CTCP! */
						handleCtcp(reply.source().get(), message);
					} else if (!channelTypes.contains(recipient.charAt(0))) {
						eventBus.post(new PrivateNoticeReceived(this, reply));
					} else {
						eventBus.post(new ChannelNoticeReceived(this, reply.source().get(), recipient, message));
					}

				/* 43x replies are for nick change errors. */
				} else if (command.equals("431")) {
					eventBus.post(new NoNicknameGivenReceived(this, reply));
				} else if (command.equals("433")) {
					if (connectionStatus == 0) {
						nickname = nicknameChooser.getNickname();
						connectionHandler.sendCommand("NICK", nickname);
					} else {
						eventBus.post(new NicknameInUseReceived(this, reply));
					}

				/* client stuff. */
				} else if (command.equalsIgnoreCase("NICK")) {
					eventBus.post(new NicknameChanged(this, reply.source().get(), parameters.get(0)));

				/* channel stuff. */
				} else if (command.equalsIgnoreCase("JOIN")) {
					eventBus.post(new ChannelJoined(this, parameters.get(0), reply.source().get()));
				} else if (command.equals("331")) {
					/* no topic is set. */
				} else if (command.equals("332")) {
					eventBus.post(new ChannelTopic(this, parameters.get(1), parameters.get(2)));
				} else if (command.equals("353")) {
					for (String nickname : parameters.get(3).split(" ")) {
						if (nickPrefixes.containsKey(nickname.substring(0, 1))) {
							nicks.add(new Nickname(nickname.substring(1), nickname.substring(0, 1)));
						} else {
							nicks.add(new Nickname(nickname, ""));
						}
					}
				} else if (command.equals("366")) {
					eventBus.post(new ChannelNicknames(this, parameters.get(1), nicks));
					nicks.clear();
				} else if (command.equalsIgnoreCase("PART")) {
					eventBus.post(new ChannelLeft(this, parameters.get(0), reply.source().get(), getOptional(parameters, 1)));
				} else if (command.equalsIgnoreCase("QUIT")) {
					eventBus.post(new ClientQuit(this, reply.source().get(), parameters.get(0)));

				} else if (channelNotJoinedHandler.willHandle(reply)) {
					channelNotJoinedHandler.handleReply(reply);

				/* basic connection housekeeping. */
				} else if (command.equalsIgnoreCase("PING")) {
					connectionHandler.sendCommand("PONG", getOptional(parameters, 0), getOptional(parameters, 1));

				} else if (connectionEstablishHandler.willHandle(reply)) {
					connectionEstablishHandler.handleReply(reply);

				/* 005 originally was a bounce message, now used to transmit useful information about the server. */
				} else if (command.equals("005")) {
					for (String parameter : parameters) {
						if (parameter.startsWith("PREFIX=")) {
							int openParen = parameter.indexOf('(');
							int closeParen = parameter.indexOf(')');
							if ((openParen != -1) && (closeParen != -1)) {
								for (int modeCharacterIndex = 1; modeCharacterIndex < (closeParen - openParen); ++modeCharacterIndex) {
									char modeCharacter = parameter.charAt(openParen + modeCharacterIndex);
									char modeSymbol = parameter.charAt(closeParen + modeCharacterIndex);
									nickPrefixes.put(String.valueOf(modeSymbol), String.valueOf(modeCharacter));
								}
								logger.debug(String.format("Parsed Prefixes: %s", nickPrefixes));
							}
						} else if (parameter.startsWith("CHANTYPES=")) {
							for (int typeIndex = 10; typeIndex < parameter.length(); ++typeIndex) {
								channelTypes.add(parameter.charAt(typeIndex));
							}
							logger.debug(String.format("Parsed Channel Types: %s", channelTypes));
						}
					}

				/* 375, 372, and 376 handle the server’s MOTD. */
				} else if (command.equals("375")) {
					/* MOTD starts. */
					motd.append(parameters.get(1)).append('\n');
				} else if (command.equals("372")) {
					motd.append(parameters.get(1)).append('\n');
				} else if (command.equals("376")) {
					motd.append(parameters.get(1)).append('\n');
					eventBus.post(new MotdReceived(this, motd.toString()));
					motd.setLength(0);

				} else if (command.equalsIgnoreCase("KICK")) {
					eventBus.post(new KickedFromChannel(this, parameters.get(0), reply.source().get(), parameters.get(1), getOptional(parameters, 2)));

				/* okay, everything else. */
				} else {
					eventBus.post(new UnknownReplyReceived(this, reply));
				}
			}
			eventBus.post(new ConnectionClosed(this));
		} catch (IOException ioe1) {
			logger.warn("I/O error", ioe1);
			eventBus.post(new ConnectionClosed(this, ioe1));
		} catch (RuntimeException re1) {
			logger.error("Runtime error", re1);
			eventBus.post(new ConnectionClosed(this, re1));
		} finally {
			established.set(false);
			eventBus.unregister(this);
			logger.info("Closing Connection.");
			try {
				Closeables.close(connectionHandler, true);
			} catch (IOException ioe1) {
				/* will not be thrown. */
			}
		}

	}

	@Subscribe
	public void connectionEstablished(ConnectionEstablished connectionEstablished) {
		if (connectionEstablished.connection() == this) {
			established.set(true);
		}
	}

	//
	// PRIVATE METHODS
	//

	/**
	 * Handles a CTCP message.
	 *
	 * @param client
	 * 		The client sending the message
	 * @param message
	 * 		The message
	 */
	private void handleCtcp(Source client, String message) {
		String[] messageWords = message.substring(1, message.length() - 1).split(" +");
		String ctcpCommand = messageWords[0];
		if (ctcpCommand.equalsIgnoreCase("DCC")) {
			if (messageWords[1].equalsIgnoreCase("SEND")) {
				Optional<InetAddress> inetAddress = parseInetAddress(messageWords[3]);
				Optional<Integer> port = Optional.fromNullable(Ints.tryParse(messageWords[4]));
				long fileSize = Optional.fromNullable(Longs.tryParse(messageWords[5])).or(-1L);
				if (inetAddress.isPresent() && port.isPresent()) {
					eventBus.post(new DccSendReceived(this, client, messageWords[2], inetAddress.get(), port.get(), fileSize));
				} else {
					logger.warn(String.format("Received malformed DCC SEND: “%s”", message));
				}
			} else if (messageWords[1].equalsIgnoreCase("ACCEPT")) {
				Optional<Integer> port = Optional.fromNullable(Ints.tryParse(messageWords[3]));
				long position = (messageWords.length > 4) ? Optional.fromNullable(Longs.tryParse(messageWords[4])).or(-1L) : -1;
				if (port.isPresent()) {
					eventBus.post(new DccAcceptReceived(this, client, messageWords[2], port.get(), position));
				} else {
					logger.warn(String.format("Received malformed DCC ACCEPT: “%s”", message));
				}
			}
		}
	}

	/**
	 * Returns an item from the list, or {@link Optional#absent()} if the list is
	 * shorter than required for the given index.
	 *
	 * @param list
	 * 		The list to get an item from
	 * @param index
	 * 		The index of the item
	 * @param <T>
	 * 		The type of the list items
	 * @return This list item wrapped in an {@link Optional}, or {@link
	 *         Optional#absent()} if the list is not long enough
	 */
	private static <T> Optional<T> getOptional(List<T> list, int index) {
		if (index < list.size()) {
			return Optional.fromNullable(list.get(index));
		}
		return Optional.absent();
	}

	/**
	 * Parses the given {@code ip} and returns an {@link InetAddress} from it.
	 *
	 * @param ip
	 * 		The IP to parse
	 * @return The parsed inet address, or {@link Optional#absent()} if no inet
	 *         address could be parsed
	 */
	private Optional<InetAddress> parseInetAddress(String ip) {
		Long ipNumber = Longs.tryParse(ip);
		if (ipNumber == null) {
			return Optional.absent();
		}

		StringBuilder hostname = new StringBuilder(15);
		hostname.append((ipNumber >>> 24) & 0xff).append('.');
		hostname.append((ipNumber >>> 16) & 0xff).append('.');
		hostname.append((ipNumber >>> 8) & 0xff).append('.');
		hostname.append(ipNumber & 0xff);
		try {
			return Optional.of(InetAddress.getByName(hostname.toString()));
		} catch (UnknownHostException uhe1) {
			return Optional.absent();
		}
	}

	/** Handles input and output for the connection. */
	private class ConnectionHandler implements Closeable {

		/** The output stream of the connection. */
		private final BandwidthCountingOutputStream outputStream;

		/** The input stream. */
		private final BandwidthCountingInputStream inputStream;

		/** The input stream of the connection. */
		private final BufferedReader inputStreamReader;

		/**
		 * Creates a new connection handler for the given input stream and output
		 * stream.
		 *
		 * @param inputStream
		 * 		The input stream of the connection
		 * @param outputStream
		 * 		The output stream of the connection
		 * @throws UnsupportedEncodingException
		 * 		if the encoding (currently “UTF-8”) is not valid
		 */
		private ConnectionHandler(InputStream inputStream, OutputStream outputStream) throws UnsupportedEncodingException {
			this.outputStream = new BandwidthCountingOutputStream(outputStream, 5, SECONDS);
			this.inputStream = new BandwidthCountingInputStream(inputStream, 5, SECONDS);
			inputStreamReader = new BufferedReader(new InputStreamReader(this.inputStream, "UTF-8"));
		}

		//
		// ACTIONS
		//

		/**
		 * Returns the current rate of the connection’s incoming side.
		 *
		 * @return The current input rate (in bytes per second)
		 */
		public long getInputRate() {
			return inputStream.getCurrentRate();
		}

		/**
		 * Returns the current rate of the connection’s outgoing side.
		 *
		 * @return The current output rate (in bytes per second)
		 */
		public long getOutputRate() {
			return outputStream.getCurrentRate();
		}

		/**
		 * Sends a command with the given parameters, skipping all {@link
		 * Optional#absent()} optionals.
		 *
		 * @param command
		 * 		The command to send
		 * @param parameters
		 * 		The parameters
		 * @throws IOException
		 * 		if an I/O error occurs
		 */
		public void sendCommand(String command, Optional<String>... parameters) throws IOException {
			List<String> setParameters = new ArrayList<String>();
			for (Optional<String> maybeSetParameter : parameters) {
				if (maybeSetParameter.isPresent()) {
					setParameters.add(maybeSetParameter.get());
				}
			}
			sendCommand(command, setParameters.toArray(new String[setParameters.size()]));
		}

		/**
		 * Sends a command with the given parameters.
		 *
		 * @param command
		 * 		The command to send
		 * @param parameters
		 * 		The parameters of the command
		 * @throws IOException
		 * 		if an I/O error occurs
		 * @throws IllegalArgumentException
		 * 		if any parameter but that last contains a space character
		 */
		public void sendCommand(String command, String... parameters) throws IOException, IllegalArgumentException {
			StringBuilder commandBuilder = new StringBuilder();

			commandBuilder.append(command);
			for (int parameterIndex = 0; parameterIndex < parameters.length; ++parameterIndex) {
				String parameter = parameters[parameterIndex];
				/* space is only allowed in the last parameter. */
				commandBuilder.append(' ');
				if (parameter.contains(" ")) {
					if (parameterIndex == (parameters.length - 1)) {
						commandBuilder.append(':');
					} else {
						throw new IllegalArgumentException(String.format("parameter “%s” must not contain space!", parameter));
					}
				}
				commandBuilder.append(parameter);
			}

			logger.trace(String.format(">> %s", commandBuilder));
			outputStream.write((commandBuilder.toString() + "\r\n").getBytes("UTF-8"));
			outputStream.flush();
		}

		/**
		 * Reads a line of reply from the connection.
		 *
		 * @return The reply
		 * @throws IOException
		 * 		if an I/O error occurs
		 * @throws EOFException
		 * 		if EOF was reached
		 */
		public Reply readReply() throws IOException, EOFException {
			String line = inputStreamReader.readLine();
			if (line == null) {
				throw new EOFException();
			}

			return Reply.parseLine(line);
		}

		//
		// CLOSEABLE METHODS
		//

		@Override
		public void close() throws IOException {
			Closeables.close(outputStream, true);
			Closeables.close(inputStreamReader, true);
			Closeables.close(inputStream, true);
		}

	}

}
