package net.pterodactylus.irc.connection;

import static java.util.Arrays.asList;

import net.pterodactylus.irc.Connection;
import net.pterodactylus.irc.Reply;
import net.pterodactylus.irc.event.ConnectionEstablished;

import com.google.common.eventbus.EventBus;

/**
 * Handles the necessary commands that establish a connection.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class ConnectionEstablishHandler implements Handler {

	private static final int FIRST_COMMAND_RECEIVED = 0x01;
	private static final int SECOND_COMMAND_RECEIVED = 0x02;
	private static final int THIRD_COMMAND_RECEIVED = 0x04;
	private static final int FOURTH_COMMAND_RECEIVED = 0x08;
	private static final int ALL_COMMANDS_RECEIVED = 0x0f;

	private final EventBus eventBus;
	private final Connection connection;
	private int connectionStatus;
	private int oldConnectionStatus;

	public ConnectionEstablishHandler(EventBus eventBus,
			Connection connection) {
		this.eventBus = eventBus;
		this.connection = connection;
	}

	@Override
	public boolean willHandle(Reply reply) {
		return asList("001", "002", "003", "004").contains(reply.command());
	}

	@Override
	public void handleReply(Reply reply) {
		String command = reply.command();
		/* replies 001-004 don’t hold information but they
		 * have to be sent on a successful connection. */
		if (command.equals("001")) {
			connectionStatus |= FIRST_COMMAND_RECEIVED;
		} else if (command.equals("002")) {
			connectionStatus |= SECOND_COMMAND_RECEIVED;
		} else if (command.equals("003")) {
			connectionStatus |= THIRD_COMMAND_RECEIVED;
		} else if (command.equals("004")) {
			connectionStatus |= FOURTH_COMMAND_RECEIVED;
		}

		if (hasConnectionStatusChanged()
				&& isFinalConnectionStatusReached()) {
			eventBus.post(new ConnectionEstablished(connection));
		}
		oldConnectionStatus = connectionStatus;
	}

	private boolean hasConnectionStatusChanged() {
		return connectionStatus != oldConnectionStatus;
	}

	private boolean isFinalConnectionStatusReached() {
		return connectionStatus == ALL_COMMANDS_RECEIVED;
	}

}
