/*
 * XdccDownloader - Download.java - Copyright © 2013 David Roden
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.pterodactylus.xdcc.data;

import static com.google.common.base.Preconditions.checkNotNull;

import java.io.OutputStream;
import java.net.InetAddress;
import java.util.Comparator;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;

import net.pterodactylus.irc.DccReceiver;

import com.google.common.base.Function;
import com.google.common.base.Optional;
import com.google.common.base.Predicate;

/**
 * Information about an ongoing download.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class Download {

	/** Predicate that identifies downloads that have started. */
	public static final Predicate<Download> FILTER_RUNNING = new Predicate<Download>() {

		@Override
		public boolean apply(Download download) {
			return download.dccReceiver() != null;
		}
	};

	/** {@link Comparator} that sorts downloads by their name. */
	public static final Comparator<Download> BY_NAME = new Comparator<Download>() {

		@Override
		public int compare(Download leftDownload, Download rightDownload) {
			return leftDownload.pack().name().compareToIgnoreCase(rightDownload.pack().name());
		}
	};

	/**  {@link Comparator} that sorts running downloads first. */
	public static final Comparator<Download> BY_RUNNING = new Comparator<Download>() {

		@Override
		public int compare(Download leftDownload, Download rightDownload) {
			return (leftDownload.dccReceiver() != null) ? -1 : (rightDownload.dccReceiver() != null) ? 1 : 0;
		}
	};

	/**
	 * Converts a download into the number of seconds left in the transfer at the
	 * current rate.
	 */
	public static final Function<Download, Optional<Long>> SECONDS_LEFT = new Function<Download, Optional<Long>>() {
		@Override
		public Optional<Long> apply(Download download) {
			DccReceiver dccReceiver = download.dccReceiver();
			if ((dccReceiver == null) || (dccReceiver.size() == -1) || (dccReceiver.currentRate() == 0)) {
				return Optional.absent();
			}
			long secondsLeft = (dccReceiver.size() - dccReceiver.progress()) / dccReceiver.currentRate();
			return Optional.of(secondsLeft);
		}
	};

	/** The bot that is being downloaded from. */
	private final Bot bot;

	/** The pack that is being downloaded. */
	private final Pack pack;

	/** The name of the file being downloaded. */
	private final AtomicReference<String> filename = new AtomicReference<String>();

	/** The size of the file being downloaded. */
	private final AtomicLong filesize = new AtomicLong();

	/** The remote address. */
	private final AtomicReference<InetAddress> remoteAddress = new AtomicReference<InetAddress>();

	/** The output stream. */
	private final AtomicReference<OutputStream> outputStream = new AtomicReference<OutputStream>();

	/** The DCC receiver. */
	private final AtomicReference<DccReceiver> dccReceiver = new AtomicReference<DccReceiver>();

	/**
	 * Creates a new download.
	 *
	 * @param bot
	 * 		The bot offering the download
	 * @param pack
	 * 		The pack being downloaded
	 */
	public Download(Bot bot, Pack pack) {
		this.bot = checkNotNull(bot, "bot must not be null");
		this.pack = checkNotNull(pack, "pack must not be null");
	}

	//
	// ACCESSORS
	//

	/**
	 * Returns the bot offering the download.
	 *
	 * @return The bot offering the download
	 */
	public Bot bot() {
		return bot;
	}

	/**
	 * The pack that is being downloaded.
	 *
	 * @return The pack being downloaded
	 */
	public Pack pack() {
		return pack;
	}

	/**
	 * The full name of the file being written.
	 *
	 * @return The full name of the file
	 */
	public String filename() {
		return filename.get();
	}

	/**
	 * Returns the size of the file.
	 *
	 * @return The size of the file
	 */
	public long filesize() {
		return filesize.get();
	}

	/**
	 * Returns the remote address to download from.
	 *
	 * @return The remote address to download from
	 */
	public InetAddress remoteAddress() {
		return remoteAddress.get();
	}

	/**
	 * The output stream that writes to the file.
	 *
	 * @return The output stream
	 */
	public OutputStream outputStream() {
		return outputStream.get();
	}

	/**
	 * The DCC receiver that is downloading the file
	 *
	 * @return The DCC receiver
	 */
	public DccReceiver dccReceiver() {
		return dccReceiver.get();
	}

	//
	// MUTATORS
	//

	/**
	 * Sets the full name of the file being downloaded.
	 *
	 * @param filename
	 * 		The full name of the file
	 * @return This download
	 */
	public Download filename(String filename) {
		this.filename.set(filename);
		return this;
	}

	/**
	 * Sets the size of the download.
	 *
	 * @param filesize
	 * 		The size of the download
	 * @return This download
	 */
	public Download filesize(long filesize) {
		this.filesize.set(filesize);
		return this;
	}

	/**
	 * Sets the remote address of the download.
	 *
	 * @param remoteAddress
	 * 		The remote address of the download
	 * @return This download
	 */
	public Download remoteAddress(InetAddress remoteAddress) {
		this.remoteAddress.set(remoteAddress);
		return this;
	}

	/**
	 * Sets the output stream the download is being written to
	 *
	 * @param outputStream
	 * 		The output stream
	 * @return This download
	 */
	public Download outputStream(OutputStream outputStream) {
		this.outputStream.set(outputStream);
		return this;
	}

	/**
	 * Sets the DCC receiver that downloads the file.
	 *
	 * @param dccReceiver
	 * 		The DCC receiver
	 * @return This download
	 */
	public Download dccReceiver(DccReceiver dccReceiver) {
		this.dccReceiver.set(dccReceiver);
		return this;
	}

	//
	// OBJECT METHODS
	//

	@Override
	public boolean equals(Object object) {
		if (!(object instanceof Download)) {
			return false;
		}
		Download download = (Download) object;
		return bot().equals(download.bot()) && pack().equals(download.pack());
	}

	@Override
	public int hashCode() {
		return bot().hashCode() ^ pack().hashCode();
	}

}
